from typing import List
import numpy
import numpy as np
import logging
from torch.utils.data import DataLoader

from hiad.utils.split_and_gather import HRImageIndex, LRPatch
from hiad.datasets.patch_dataset import PatchDataset
from hiad.datasets.syn_dataset import AnomalySynDataset
from hiad.syn import *
from .base import BaseDetector
from .destseg.model.destseg import TeacherNet, StudentNet, SegmentationNet
from .destseg.model.model_utils import l2_normalize
from .destseg.model.losses import *


class HRDesTSeg(BaseDetector):

    def __init__(self,
                 patch_size: int,  # base
                 DTD,
                 steps,
                 de_st_steps,
                 logger: logging.Logger,  # base
                 device: torch.device,  # base
                 dest = True,
                 lr_de_st = 0.4,
                 lr_res = 0.1,
                 lr_seghead = 0.01,
                 eval_per_steps = 500,
                 log_per_steps = 50,
                 gamma = 4,
                 seed: int = 0,  #base
                 fusion_weights = None,
                 early_stop_epochs=-1,
                 **kwargs):

        super().__init__(patch_size, device, fusion_weights, logger, seed, early_stop_epochs)

        self.max_anomaly_score = None
        self.min_anomaly_score = None
        self.steps = steps
        self.de_st_steps =  de_st_steps

        self.lr_de_st = lr_de_st
        self.lr_res = lr_res
        self.lr_seghead = lr_seghead

        self.eval_per_steps = eval_per_steps
        self.log_per_steps = log_per_steps
        self.gamma = gamma
        self.DTD = DTD
        self.dest = dest
        self.teacher_net = TeacherNet().to(self.device)
        self.student_net = StudentNet().to(self.device)
        self.segmentation_net = SegmentationNet(inplanes = 384).to(self.device)


    def create_dataset(self, patches: List[LRPatch], training: bool, task_name: str ):
        if not training:
            dataset = PatchDataset(patches=patches, training=training, task_name=task_name)
        else:
            dataset = AnomalySynDataset(patches=patches, synthesizer = ImageBlendingSynthesizer(p=1.0, anomaly_source = self.DTD),
                                        training=training, task_name = task_name)
        return dataset

    def destseg_embedding(self, input_tensor: torch.Tensor, model):
        return model(input_tensor)

    @torch.no_grad()
    def embedding(self, input_tensor: torch.Tensor ) -> List[torch.Tensor]:
        return self.destseg_embedding(input_tensor, self.teacher_net)

    @torch.no_grad()
    def get_multi_resolution_fusion_embeddings(self, data) -> List[torch.Tensor]:
        return self.destseg_multi_resolution_fusion_embeddings(data, self.teacher_net, gt_image=False)

    def destseg_multi_resolution_fusion_embeddings(self, data, model, gt_image = False) -> List[torch.Tensor]:
        image = data['gt_image'].to(self.device) if gt_image else data['image'].to(self.device)
        embeddings = self.destseg_embedding(image, model)
        low_resolution_image_keys = [key for key in data if key.startswith('gt_low_resolution_image' if gt_image else 'low_resolution_image')]
        if len(low_resolution_image_keys) == 0:
            return embeddings
        else:
            if self.fusion_weights is not None:
                assert len(self.fusion_weights) == len(low_resolution_image_keys) + 1, "fusion_weights must be the same length as ds_factor."
                fusion_weights = [weight / sum(self.fusion_weights) for weight in self.fusion_weights]
            else:
                fusion_weights = [1 / (len(low_resolution_image_keys) + 1)] * (len(low_resolution_image_keys) + 1)

            embeddings = [[embedding * fusion_weights[0]] for embedding in embeddings]

            low_resolution_image_keys.sort(key=lambda item: int(item.split('_')[-1]))

            for rs_index, low_resolution_image_key in enumerate(low_resolution_image_keys):
                low_resolution_images = data[low_resolution_image_key].to(self.device)
                low_resolution_indexes = data[low_resolution_image_key.replace('image', 'index')]
                low_resolution_embeddings = self.destseg_embedding(low_resolution_images, model)

                for i, low_resolution_embedding in enumerate(low_resolution_embeddings):
                    downsampling_embedding = []
                    for feature, index in zip(low_resolution_embedding, low_resolution_indexes):
                        stride_H, stride_W = self.patch_size[1] / feature.shape[1], self.patch_size[0] / feature.shape[2]
                        index = HRImageIndex.from_str(index)
                        x_start = index.x / stride_W
                        y_start = index.y / stride_H
                        x_end = x_start + index.width / stride_W
                        y_end = y_start + index.height / stride_H
                        downsampling_embedding.append(feature[:, int(y_start):int(y_end), int(x_start):int(x_end)])

                    try:
                        downsampling_embedding = torch.stack(downsampling_embedding)
                    except:
                        first_embedding = downsampling_embedding[0]
                        downsampling_embedding = [
                            F.interpolate(feat.unsqueeze(0), size=(first_embedding.shape[-2], first_embedding.shape[-1]),
                                          mode='bilinear').squeeze(0)
                            for feat in downsampling_embedding[1:]]
                        downsampling_embedding = [first_embedding] + downsampling_embedding
                        downsampling_embedding = torch.stack(downsampling_embedding)

                    embeddings[i].append(fusion_weights[rs_index+1] * F.interpolate(
                        downsampling_embedding,
                        size=(embeddings[i][-1].shape[-2], embeddings[i][-1].shape[-1]),
                        mode="bilinear",
                        align_corners=False,
                    ))
            embeddings = [torch.sum(torch.stack(embedding), dim=0, keepdim=False) for embedding in embeddings]
            return embeddings


    def train_step(self,
                   train_dataloader: DataLoader,
                   task_name: str,
                   checkpoint_path: str,
                   val_dataloader: DataLoader = None,
                   evaluators = None,
                   ) -> bool:

        seg_optimizer = torch.optim.SGD(
            [
                {"params": self.segmentation_net.res.parameters(), "lr": self.lr_res},
                {"params": self.segmentation_net.head.parameters(), "lr": self.lr_seghead},
            ],
            lr=0.001,
            momentum=0.9,
            weight_decay=1e-4,
            nesterov=False,
        )

        de_st_optimizer = torch.optim.SGD(
            [
                {"params": self.student_net.parameters(), "lr": self.lr_de_st},
            ],
            lr=0.4,
            momentum=0.9,
            weight_decay=1e-4,
            nesterov=False,
        )
        global_step = 0
        best_metrics = {}

        while global_step <= self.steps:
            torch.cuda.empty_cache()
            for _, sample_batched in enumerate(train_dataloader):
                seg_optimizer.zero_grad()
                de_st_optimizer.zero_grad()

                if global_step < self.de_st_steps:
                    self.student_net.train()
                    self.segmentation_net.eval()
                else:
                    self.student_net.eval()
                    self.segmentation_net.train()

                outputs_teacher_aug = [
                    l2_normalize(output_t.detach()) for output_t in self.destseg_multi_resolution_fusion_embeddings(sample_batched,
                                                                                                          self.teacher_net,
                                                                                                          gt_image = False)
                ]

                outputs_student_aug = [
                    l2_normalize(output_s) for output_s in self.destseg_multi_resolution_fusion_embeddings(sample_batched,
                                                                                                 self.student_net,
                                                                                                gt_image =  False)
                ]

                output = torch.cat(
                    [
                        F.interpolate(
                            -output_t * output_s,
                            size=outputs_student_aug[0].size()[2:],
                            mode="bilinear",
                            align_corners=False,
                        )
                        for output_t, output_s in zip(outputs_teacher_aug, outputs_student_aug)
                    ],
                    dim=1,
                )

                output_segmentation = self.segmentation_net(output)

                if self.dest:
                    outputs_student = outputs_student_aug
                else:
                    outputs_student = [
                        l2_normalize(output_s) for output_s in self.destseg_multi_resolution_fusion_embeddings(sample_batched, self.student_net,
                                                                                                     gt_image=True)
                    ]

                outputs_teacher = [
                    l2_normalize(output_t.detach()) for output_t in self.destseg_multi_resolution_fusion_embeddings(sample_batched, self.teacher_net,
                                                                                                     gt_image = True)
                ]

                output_de_st_list = []
                for output_t, output_s in zip(outputs_teacher, outputs_student):
                    a_map = 1 - torch.sum(output_s * output_t, dim=1, keepdim=True)
                    output_de_st_list.append(a_map)

                mask = F.interpolate(
                    sample_batched['mask'].unsqueeze(1),
                    size=output_segmentation.size()[2:],
                    mode="bilinear",
                    align_corners=False,
                ).to(self.device)

                mask = torch.where(
                    mask < 0.5, torch.zeros_like(mask), torch.ones_like(mask)
                )
                cosine_loss_val = cosine_similarity_loss(output_de_st_list)
                focal_loss_val = focal_loss(output_segmentation, mask, gamma=self.gamma)
                l1_loss_val = l1_loss(output_segmentation, mask)

                if global_step < self.de_st_steps:
                    total_loss_val = cosine_loss_val
                    total_loss_val.backward()
                    de_st_optimizer.step()
                else:
                    total_loss_val = focal_loss_val + l1_loss_val
                    total_loss_val.backward()
                    seg_optimizer.step()

                global_step += 1
                if global_step % self.log_per_steps == 0:
                    if global_step < self.de_st_steps:
                        self.logger.info(
                            f"Training at global step {global_step}/{self.steps}, cosine loss: {round(float(cosine_loss_val), 4)}"
                        )
                    else:
                        self.logger.info(
                            f"Training at global step {global_step}/{self.steps}, focal loss: {round(float(focal_loss_val), 4)}, l1 loss: {round(float(l1_loss_val), 4)}"
                        )

                if global_step >= self.de_st_steps and global_step % self.eval_per_steps == 0 and val_dataloader is not None:
                    best_metrics = self.val_step(val_dataloader, evaluators, checkpoint_path, best_metrics)
                    if self.early_stop_cur == 0:
                        return True

        return True if best_metrics!={} else False


    @torch.no_grad()
    def inference_step(self,
                   test_dataloader: DataLoader,
                   task_name: str,
                   )-> List[numpy.ndarray]:

        self.teacher_net.eval()
        self.student_net.eval()
        self.segmentation_net.eval()

        preds = []
        with torch.no_grad():
            for _, sample_batched in enumerate(test_dataloader):

                outputs_teacher_aug = [
                    l2_normalize(output_t.detach()) for output_t in self.destseg_multi_resolution_fusion_embeddings(sample_batched,
                                                                                                          self.teacher_net,
                                                                                                          gt_image=False)
                ]

                outputs_student_aug = [
                    l2_normalize(output_s) for output_s in self.destseg_multi_resolution_fusion_embeddings(sample_batched,
                                                                                                 self.student_net,
                                                                                                 gt_image=False)
                ]

                output = torch.cat(
                    [
                        F.interpolate(
                            -output_t * output_s,
                            size=outputs_student_aug[0].size()[2:],
                            mode="bilinear",
                            align_corners=False,
                        )
                        for output_t, output_s in zip(outputs_teacher_aug, outputs_student_aug)
                    ],
                    dim=1,
                )

                output_segmentation = self.segmentation_net(output)

                preds.append(F.interpolate(
                    output_segmentation,
                    size = sample_batched['mask'].size()[-2:],
                    mode="bilinear",
                    align_corners=False,
                ).squeeze(1).cpu().numpy())

        preds = np.concatenate(preds, axis=0)
        if task_name != 'val':
            preds = self.patch_post_processing(preds)
        preds = [pred for pred in preds]
        return preds


    def save_checkpoint(self, checkpoint_path: str):
        torch.save({'teacher': self.teacher_net.state_dict(),
                    'student': self.student_net.state_dict(),
                    'segmentation': self.segmentation_net.state_dict(),
                    'max_anomaly_score': self.max_anomaly_score,
                    'min_anomaly_score': self.min_anomaly_score,
                    'fusion_weights': self.fusion_weights},
                   checkpoint_path)


    def load_checkpoint(self, checkpoint_path: str):
        state_dict = torch.load(checkpoint_path, map_location=self.device)
        self.teacher_net.load_state_dict(state_dict['teacher'])
        self.student_net.load_state_dict(state_dict['student'])
        self.segmentation_net.load_state_dict(state_dict['segmentation'])
        self.min_anomaly_score = state_dict['min_anomaly_score']
        self.max_anomaly_score = state_dict['max_anomaly_score']
        self.fusion_weights = state_dict['fusion_weights'] if 'fusion_weights' in state_dict else None


    @staticmethod
    def get_image_score(segmentations, T = 100, batch_size = 32):
        preds = torch.tensor(segmentations)
        preds = preds.split(batch_size, dim=0)
        image_scores = []
        for pred in preds:
            image_score, _ = torch.sort(
                pred.view(pred.size(0), -1).cuda(),
                dim=1,
                descending=True,
            )
            image_score = torch.mean(
                image_score[:, : T], dim=1
            )
            image_scores.append(image_score.cpu().numpy())

        image_scores = np.concatenate(image_scores, axis=0)
        return image_scores


from abc import ABC, abstractmethod
from typing import List, Dict
import numpy as np
import torch
import tqdm
import numpy
import logging
from sklearn.cluster import DBSCAN, KMeans
import torch.nn.functional as F
import os
from PIL import Image
from easydict import EasyDict

from hiad.task.utils import patch_embedding
from hiad.utils.split_and_gather import HRSample, HRImage
from hiad.utils.split_and_gather import MultiResolutionIndex
from hiad.utils.base import ImageIndexes2NdArray


class BaseTaskGenerator(ABC):
    r"""
       The TaskGenerator divides image patches into different tasks for processing,
         with each task assigned to a detector. This class serves as the base class for all TaskGenerator implementations

       Args:
           deviec (List[int]): GPU ids.
    """

    def __init__(self, device = None):
        self.device = device

    @abstractmethod
    def create_tasks(self,
                        samples: List[HRSample],
                        indexes: List[MultiResolutionIndex],
                        **kwargs) -> List[Dict]:
        r"""
           The core method for task splitting
           Args:
               samples (List[HRSample]): The provided image set, typically the training set.
               indexes (List[MultiResolutionIndex]): indexes set.
           returns:
               Returns the list of created tasks. List(dict)
        """
        raise NotImplementedError


class O2OTaskGenerator(BaseTaskGenerator):
    r"""
        Assign all indexes to a single task.
    """

    def create_tasks(self,
                        samples: List[HRSample],
                        indexes: List[MultiResolutionIndex],
                        **kwargs) -> List[Dict]:

        return [{"name": f'task_{i}', "indexes": [index]} for i, index in enumerate(indexes)]


class A2OTaskGenerator(BaseTaskGenerator):
    r"""
          Assign each index to a separate task.
    """
    def create_tasks(self,
                        samples: List[HRSample],
                        indexes: List[MultiResolutionIndex],
                        **kwargs) -> List[Dict]:
        return [{"name": f'task_0', "indexes": indexes}]


class NeighborhoodTaskGenerator(BaseTaskGenerator):
    r"""
        Assign all indexes within a small spatial neighborhood to a single task.

        Args:
            num_groups_width (int): Number of neighborhoods divided along the image width. if -1, will be set to the number of patches.
            num_groups_height (int): Number of neighborhoods divided along the image height. if -1, will be set to the number of patches.
                if None, will be set to the `num_groups_width`.
            deviec (List[int]): GPU ids.
     """

    def __init__(self, num_groups_width, num_groups_height = None, device=None):
        assert num_groups_width is not None
        self.num_groups_row = num_groups_width
        if num_groups_height is None:
            self.num_groups_column = self.num_groups_row
        else:
            self.num_groups_column = num_groups_height
        assert (self.num_groups_row > 0 or self.num_groups_row == -1) and (self.num_groups_column > 0 or self.num_groups_column == -1)
        super().__init__(device)

    def create_tasks(self,
                        samples: List[HRSample],
                        indexes: List[MultiResolutionIndex],
                        **kwargs) -> List[Dict]:

        indexes_array = ImageIndexes2NdArray(indexes)

        if self.num_groups_row == -1:
            num_row = indexes_array.shape[0]
        else:
            num_row = self.num_groups_row

        if self.num_groups_column == -1:
            num_column = indexes_array.shape[1]
        else:
            num_column = self.num_groups_column

        def split_array(arr, m, n):
            row_splits = numpy.array_split(arr, m, axis=0)
            result = [numpy.array_split(row, n, axis=1) for row in row_splits]
            return [submat.flatten() for row in result for submat in row]

        tasks = split_array(indexes_array,num_column, num_row)
        tasks = [task for task in tasks if len(task) != 0]
        tasks = [{'name': f'task_{index}', 'indexes': task} for index, task in enumerate(tasks)]
        return tasks


class SpatialClusteringTaskGenerator(BaseTaskGenerator):
    r"""
        Cluster spatial patches with high visual similarity.

        Args:
            detector_class (class): The provided detector class template, used to create detector instances across multiple processes.
                    Supported detector types are listed in `detectors.__init__`.
            detector_config: Detector configuration parameters.
            batch_size (int): Feature extraction Batch Size.
            cluster_number (int): Number of clusters.
            feature_resolution (int): The extracted multi-scale features are uniformly downsampled to a (feature_resolution * feature_resolution) resolution for clustering.
            fusion_weights (list): Weights used for feature fusion. If None, a set of equal weights will be used.
            deviec (List[int]): GPU ids.
     """

    def __init__(self,
                 detector_class,
                 detector_config,
                 batch_size,
                 cluster_number = 4,
                 feature_resolution = 1, fusion_weights = None, device = None):

        self.detector_class = detector_class

        if type(detector_config) == dict:
            detector_config = EasyDict(detector_config)

        self.detector_config = detector_config
        self.batch_size = batch_size
        self.cluster_number = cluster_number
        if isinstance(feature_resolution, int):
            self.feature_resolution = [feature_resolution, feature_resolution]
        else:
            self.feature_resolution = feature_resolution
        self.fusion_weights = fusion_weights
        super().__init__(device)

    def create_tasks(self,
                        samples: List[HRSample],
                        indexes: List[MultiResolutionIndex],
                        **kwargs) -> List[Dict]:

        main_logger = logging.getLogger('main')
        main_logger.info(f'Start patch clustering, devices: {self.device}')

        embedding_pairs = patch_embedding(samples,
                                          indexes,
                                          self.detector_class,
                                          self.detector_config.patch,
                                          self.batch_size,
                                          self.feature_resolution,
                                          self.fusion_weights,
                                          self.device)

        index_embeddings = {}
        for _, index, embedding in embedding_pairs:
            if index not in index_embeddings:
                index_embeddings[index]= []
            index_embeddings[index].append(embedding)

        for index in index_embeddings:
            index_embeddings[index] = np.mean(np.stack(index_embeddings[index]),axis=0).reshape(-1)

        embeddings = np.stack([index_embeddings[key] for key in index_embeddings])
        embeddings = F.normalize(torch.tensor(embeddings), p=2, dim=1).numpy()

        indexes = [ key for key in index_embeddings]

        clustering = KMeans(n_clusters=self.cluster_number, random_state=1)
        labels = clustering.fit_predict(embeddings)
        tasks = {}
        for index, label in zip(indexes, labels):
            if f'task_{label}' not in tasks:
                tasks[f'task_{label}'] = []
            tasks[f'task_{label}'].append(index)

        tasks = [{'name': task_name, 'indexes': tasks[task_name]} for task_name in tasks]
        tasks.sort(key= lambda item: int(item['name'].split('_')[-1]))
        return tasks


class ThresholdClusteringTaskGenerator(BaseTaskGenerator):
    r"""
        Cluster spatial patches with high visual similarity. Unlike `SpatialClusteringTaskGenerator`,
        this class does not require specifying the number of clusters in advance. Clustering is controlled by `cluster_threshold`:
        indexes with similarity ≥ cluster_threshold are grouped into the same cluster.
     """

    def __init__(self, detector_class, detector_config, batch_size,
                       cluster_threshold = 0.5, feature_resolution = 1, fusion_weights = None, device = None):

        self.detector_class = detector_class

        if type(detector_config) == dict:
            detector_config = EasyDict(detector_config)

        self.detector_config = detector_config
        self.batch_size = batch_size
        self.cluster_threshold = cluster_threshold
        if isinstance(feature_resolution, int):
            self.feature_resolution = [feature_resolution, feature_resolution]
        else:
            self.feature_resolution = feature_resolution
        self.fusion_weights = fusion_weights
        super().__init__(device)

    def create_tasks(self,
                        samples: List[HRSample],
                        indexes: List[MultiResolutionIndex],
                        **kwargs) -> List[Dict]:

        main_logger = logging.getLogger('main')
        main_logger.info(f'Start patch clustering, devices: {self.device}')

        embedding_pairs = patch_embedding(samples,
                                          indexes,
                                          self.detector_class,
                                          self.detector_config.patch,
                                          self.batch_size,
                                          self.feature_resolution,
                                          self.fusion_weights,
                                          self.device)

        index_embeddings = {}
        for _, index, embedding in embedding_pairs:
            if index not in index_embeddings:
                index_embeddings[index] = []
            index_embeddings[index].append(embedding)

        for index in index_embeddings:
            index_embeddings[index] = np.mean(np.stack(index_embeddings[index]), axis=0).reshape(-1)

        embeddings = np.stack([index_embeddings[key] for key in index_embeddings])
        embeddings = F.normalize(torch.tensor(embeddings),p=2,dim=1).numpy()

        indexes = [key for key in index_embeddings]

        clustering = DBSCAN(eps=self.cluster_threshold, min_samples=1)
        labels = clustering.fit_predict(embeddings)

        tasks = {}
        noise_tasks = []
        for index, label in zip(indexes, labels):
            if label == -1:
                noise_tasks.append(index)
            else:
                if f'task_{label}' not in tasks:
                    tasks[f'task_{label}'] = []
                tasks[f'task_{label}'].append(index)
        for noise_task in noise_tasks:
            tasks[f'task_{len(tasks)}'] = [noise_task]
        tasks = [{'name': task_name, 'indexes': tasks[task_name]} for task_name in tasks]
        tasks.sort(key= lambda item: int(item['name'].split('_')[-1]))
        return tasks


class RetrieveTaskGenerator(BaseTaskGenerator):
    r"""
        Clustering is based solely on patch similarity, ignoring spatial location. This class performs dynamic retrieval during inference.

        Args:
            detector_class (class): The provided detector class template, used to create detector instances across multiple processes.
                    Supported detector types are listed in `detectors.__init__`.
            detector_config: Detector configuration parameters.
            batch_size (int): Feature extraction Batch Size.
            cluster_number (int): Number of clusters.
            feature_resolution (int): The extracted multi-scale features are uniformly downsampled to a (feature_resolution * feature_resolution) resolution for clustering.
            save_root (str): If provided, the clustering results of the patches will be saved.
            fusion_weights (list): Weights used for feature fusion. If None, a set of equal weights will be used.
            deviec (List[int]): GPU ids.
     """
    def __init__(self, detector_class, detector_config, batch_size,
                       cluster_number = 4, feature_resolution = 1, save_root = None, fusion_weights = None, device = None):

        self.detector_class = detector_class
        if type(detector_config) == dict:
            detector_config = EasyDict(detector_config)

        self.detector_config = detector_config
        self.batch_size = batch_size
        self.cluster_number = cluster_number
        self.feature_resolution = feature_resolution
        self.save_root = save_root
        if isinstance(feature_resolution, int):
            self.feature_resolution = [feature_resolution, feature_resolution]
        else:
            self.feature_resolution = feature_resolution
        self.fusion_weights = fusion_weights
        super().__init__(device)


    def create_tasks(self,
                        samples: List[HRSample],
                        indexes: List[MultiResolutionIndex],
                        **kwargs) -> List[Dict]:

        main_logger = logging.getLogger('main')
        main_logger.info(f'Start Retrieve clustering, devices: {self.device}')

        embedding_pairs = patch_embedding(samples,
                                        indexes,
                                        self.detector_class,
                                        self.detector_config.patch,
                                        self.batch_size,
                                        self.feature_resolution,
                                        self.fusion_weights,
                                        self.device)

        embeddings = [pair[-1].reshape(-1) for pair in embedding_pairs]
        paths_and_indexes = [ pair[:-1] for pair in embedding_pairs]

        embeddings = np.stack(embeddings)
        embeddings = F.normalize(torch.tensor(embeddings), p=2, dim=1).numpy()

        clustering = KMeans(n_clusters=self.cluster_number, random_state=1)
        labels = clustering.fit_predict(embeddings)

        centers = { f'retrieve_task_{label}': center for label, center in enumerate(clustering.cluster_centers_)}

        tasks = {}
        for path_and_index, label in zip(paths_and_indexes, labels):
            if f'retrieve_task_{label}' not in tasks:
                tasks[f'retrieve_task_{label}'] = []
            tasks[f'retrieve_task_{label}'].append(path_and_index)

        tasks = [{'name': task_name, 'paths_and_indexes': tasks[task_name],
                  'center': centers[task_name], 'feature_resolution':self.feature_resolution, 'fusion_weights': self.fusion_weights } for task_name in tasks]
        tasks.sort(key = lambda item: int(item['name'].split('_')[-1]))

        if self.save_root is not None:
            for task in tasks:
                task_root = os.path.join(self.save_root, task['name'])
                os.makedirs(task_root, exist_ok=True)
                for i, (path, index) in enumerate(tqdm.tqdm(task['paths_and_indexes'], desc=f'Save retrieve patch {task["name"]}')):
                    image = HRImage(image_path=path, is_mask=False, image_size = kwargs.get('image_size', None))
                    image.open()
                    patch = image[index.main_index]
                    image.close()
                    Image.fromarray(patch).save(os.path.join(task_root,f'patch_{i}.jpg'))
        return tasks


import torch


def get_optimizer(parameters, config):
    if config.type == "AdamW":
        return torch.optim.AdamW(parameters, **config.kwargs)
    elif config.type == "Adam":
        return torch.optim.Adam(parameters, **config.kwargs)
    elif config.type == "SGD":
        return torch.optim.SGD(parameters, **config.kwargs)
    else:
        raise NotImplementedError



import logging
from typing import List
import torch
import numpy
from torch.utils.data import DataLoader
import torch.nn.functional as F

from hiad.datasets.patch_dataset import PatchDataset
from hiad.datasets.syn_dataset import AnomalySynDataset
from hiad.utils.split_and_gather import LRPatch,HRImageIndex
from hiad.syn import *
from .realnet.models.backbones import Backbone
from .realnet.models.model_helper import ModelHelper, to_device
from .realnet.utils.misc_helper import summary_model, AverageMeter
from .realnet.utils.optimizer_helper import get_optimizer
from .realnet.utils.criterion_helper import build_criterion
from .base import BaseDetector


class HRRealNet(BaseDetector):
    def __init__(self,
                 patch_size: int, # base
                 synthesizer,
                 structure,
                 net,
                 trainer,
                 criterion,
                 logger: logging.Logger, #base
                 device: torch.device, #base
                 seed: int = 0, #base
                 fusion_weights = None,
                 early_stop_epochs = -1,
                 **kwargs):
        super().__init__(patch_size, device, fusion_weights, logger, seed, early_stop_epochs)

        self.synthesizer = synthesizer
        self.structure = structure
        self.net = net
        self.trainer = trainer
        self.criterion = criterion

        self.max_anomaly_score = None
        self.min_anomaly_score = None

        layers = []
        for block in self.structure:
            layers.extend([layer.idx for layer in block.layers])
        layers = list(set(layers))
        layers.sort()

        self.net[0].kwargs['outlayers'] = layers
        self.net[1].kwargs = self.net[1].get('kwargs', {})
        self.net[1].kwargs['structure'] = self.structure

        self.synthesizer = eval(self.synthesizer.name)(**self.synthesizer.kwargs)
        self.backbone = Backbone(**self.net[0].kwargs)
        self.backbone.to(self.device)
        self.backbone.eval()

        self.net = self.net[1:]
        if 'prev' in self.net[0]:
            del self.net[0]['prev']
            kwargs = self.net[0].get('kwargs',{})
            kwargs['inplanes'] = self.backbone.get_outplanes()
            kwargs['instrides'] = self.backbone.get_outstrides()
            self.net[0].kwargs = kwargs

        self.realnet = ModelHelper(self.net, self.device)
        self.realnet.to(self.device)


    def create_dataset(self, patches: List[LRPatch], training: bool, task_name: str ):
        if not training:
            dataset = PatchDataset(patches=patches, training=training, task_name=task_name)
        else:
            dataset = AnomalySynDataset(patches=patches, synthesizer=self.synthesizer, training=training, task_name = task_name)
        return dataset


    @torch.no_grad()
    def embedding(self, input_tensor, train = False):
        input_tensor = input_tensor.to(self.device)
        embeddings = self.backbone(input_tensor, train=train)
        return embeddings

    def realnet_emb(self, data, train = False):
        feats_dict = {}

        if train:
            feats, gt_feats = self.get_multi_resolution_fusion_embeddings(data, train)

            feats = {self.backbone.outlayers[idx]: {
                "feat": feats[idx],
                "stride": self.backbone.layers_strides[self.backbone.outlayers[idx]],
                "planes": self.backbone.layers_planes[self.backbone.outlayers[idx]],
            } for idx in range(len(self.backbone.outlayers))}

            feats_dict.update({"feats": feats})

            gt_feats = {self.backbone.outlayers[idx]: {
                "feat": gt_feats[idx],
                "stride": self.backbone.layers_strides[self.backbone.outlayers[idx]],
                "planes": self.backbone.layers_planes[self.backbone.outlayers[idx]],
            } for idx in range(len(self.backbone.outlayers))}
            feats_dict.update({"gt_feats": gt_feats})
        else:
            feats = self.get_multi_resolution_fusion_embeddings(data, train)

            feats = {self.backbone.outlayers[idx]: {
                "feat": feats[idx],
                "stride": self.backbone.layers_strides[self.backbone.outlayers[idx]],
                "planes": self.backbone.layers_planes[self.backbone.outlayers[idx]],
            } for idx in range(len(self.backbone.outlayers))}

            feats_dict.update({"feats": feats})
        return feats_dict


    @torch.no_grad()
    def get_multi_resolution_fusion_embeddings(self, data, train = False):
        data = to_device(data, device=self.device)
        image = data["image"]
        feats = self.embedding(image)
        low_resolution_image_keys = [key for key in data if key.startswith('low_resolution_image')]

        if len(low_resolution_image_keys) != 0:
            if self.fusion_weights is not None:
                assert len(self.fusion_weights) == len(low_resolution_image_keys) + 1
                fusion_weights = [weight / sum(self.fusion_weights) for weight in self.fusion_weights]
            else:
                fusion_weights = [1 / (len(low_resolution_image_keys) + 1)] * (len(low_resolution_image_keys) + 1)

            feats = [[feat * fusion_weights[0]] for feat in feats]
            low_resolution_image_keys.sort(key=lambda item: int(item.split('_')[-1]))
            for rs_index, low_resolution_image_key in enumerate(low_resolution_image_keys):
                low_resolution_images = data[low_resolution_image_key]
                low_resolution_indexes = data[low_resolution_image_key.replace('image', 'index')]
                low_resolution_embeddings = self.embedding(low_resolution_images)
                for i, low_resolution_embedding in enumerate(low_resolution_embeddings):
                    downsampling_embedding = []
                    for feature, index in zip(low_resolution_embedding, low_resolution_indexes):
                        stride_H, stride_W = self.patch_size[1] / feature.shape[1], self.patch_size[0] / feature.shape[2]
                        index = HRImageIndex.from_str(index)
                        x_start = index.x / stride_W
                        y_start = index.y / stride_H
                        x_end = x_start + index.width / stride_W
                        y_end = y_start + index.height / stride_H
                        downsampling_embedding.append(feature[:, int(y_start):int(y_end), int(x_start):int(x_end)])

                    try:
                        downsampling_embedding = torch.stack(downsampling_embedding)
                    except:
                        first_embedding = downsampling_embedding[0]
                        downsampling_embedding = [
                            F.interpolate(feat.unsqueeze(0), size=(first_embedding.shape[-2], first_embedding.shape[-1]),
                                          mode='bilinear').squeeze(0)
                            for feat in downsampling_embedding[1:]]
                        downsampling_embedding = [first_embedding] + downsampling_embedding
                        downsampling_embedding = torch.stack(downsampling_embedding)

                    feats[i].append(fusion_weights[rs_index+1] * F.interpolate(
                        downsampling_embedding,
                        size=(feats[i][-1].shape[-2], feats[i][-1].shape[-1]),
                        mode="bilinear",
                        align_corners=False,
                    ))

            feats = [torch.sum(torch.stack(feat), dim=0, keepdim=False) for feat in feats]

        if train:
            gt_image = data["gt_image"]
            gt_feats = self.embedding(gt_image)
            gt_low_resolution_image_keys = [key for key in data if key.startswith('gt_low_resolution_image')]

            if len(gt_low_resolution_image_keys) != 0:
                if self.fusion_weights is not None:
                    assert len(self.fusion_weights) == len(low_resolution_image_keys) + 1
                    fusion_weights = [weight / sum(self.fusion_weights) for weight in self.fusion_weights]
                else:
                    fusion_weights = [1 / (len(low_resolution_image_keys) + 1)] * (len(low_resolution_image_keys) + 1)

                gt_feats = [[feat * fusion_weights[0]] for feat in gt_feats]
                gt_low_resolution_image_keys.sort(key=lambda item: int(item.split('_')[-1]))

                for rs_index, gt_low_resolution_image_key in enumerate(gt_low_resolution_image_keys):
                    gt_low_resolution_images = data[gt_low_resolution_image_key]
                    gt_low_resolution_indexes = data[gt_low_resolution_image_key.replace('image', 'index')]
                    gt_low_resolution_embeddings = self.embedding(gt_low_resolution_images)

                    for i, gt_low_resolution_embedding in enumerate(gt_low_resolution_embeddings):
                        downsampling_embedding = []
                        for feature, index in zip(gt_low_resolution_embedding, gt_low_resolution_indexes):
                            stride_H, stride_W = self.patch_size[1] / feature.shape[1], self.patch_size[0] / feature.shape[2]
                            index = HRImageIndex.from_str(index)
                            x_start = index.x / stride_W
                            y_start = index.y / stride_H
                            x_end = x_start + index.width / stride_W
                            y_end = y_start + index.height / stride_H
                            downsampling_embedding.append(feature[:, int(y_start):int(y_end), int(x_start):int(x_end)])

                        try:
                            downsampling_embedding = torch.stack(downsampling_embedding)
                        except:
                            first_embedding = downsampling_embedding[0]
                            downsampling_embedding = [
                                F.interpolate(feat.unsqueeze(0),
                                              size=(first_embedding.shape[-2], first_embedding.shape[-1]),
                                              mode='bilinear').squeeze(0)
                                for feat in downsampling_embedding[1:]]
                            downsampling_embedding = [first_embedding] + downsampling_embedding
                            downsampling_embedding = torch.stack(downsampling_embedding)

                        gt_feats[i].append(fusion_weights[rs_index+1] * F.interpolate(
                            downsampling_embedding,
                            size=(gt_feats[i][-1].shape[-2], gt_feats[i][-1].shape[-1]),
                            mode="bilinear",
                            align_corners=False,
                        ))
                gt_feats = [torch.sum(torch.stack(feat), dim=0, keepdim=False) for feat in gt_feats]

            return feats, gt_feats

        return feats


    def train_step(self,
                   train_dataloader: DataLoader,
                   task_name: str,
                   checkpoint_path: str,
                   val_dataloader: DataLoader = None,
                   evaluators = None,
                   ) -> bool:

        self.realnet.afs.init_idxs(self, train_dataloader)

        summary_model(self.realnet, self.logger)
        layers = []

        for module in self.net:
            layers.append(module["name"])

        frozen_layers = self.realnet.frozen_layers
        active_layers = [layer for layer in layers if layer not in frozen_layers]

        self.logger.info("layers: {}".format(layers))
        self.logger.info("frozen layers: {}".format(frozen_layers))
        self.logger.info("active layers: {}".format(active_layers))

        parameters = [
            {"params": getattr(self.realnet, layer).parameters()} for layer in active_layers
        ]
        optimizer = get_optimizer(parameters, self.trainer.optimizer)

        best_metrics = {}
        last_epoch = 0

        criterion = build_criterion(self.criterion)

        for epoch in range(last_epoch, self.trainer.max_epoch):

            last_iter = epoch * len(train_dataloader)
            losses = AverageMeter(self.trainer.log_per_steps)

            for i, input in enumerate(train_dataloader):
                torch.cuda.empty_cache()
                self.realnet.train()
                curr_step = last_iter + i + 1

                outputs = self.realnet_emb(input, train=True)
                outputs.update(input)
                outputs = self.realnet(outputs, train =True)

                loss = []
                for name, criterion_loss in criterion.items():
                    weight = criterion_loss.weight
                    loss.append(weight * criterion_loss(outputs))

                loss = torch.sum(torch.stack(loss))
                losses.update(loss.item())
                optimizer.zero_grad()
                loss.backward()
                if self.trainer.get("clip_max_norm", None):
                    max_norm = self.trainer.clip_max_norm
                    torch.nn.utils.clip_grad_norm_(self.realnet.parameters(), max_norm)
                optimizer.step()

                if curr_step % self.trainer.log_per_steps == 0:
                    self.logger.info(
                            "Epoch: [{0}/{1}]\t"
                            "Iter: [{2}/{3}]\t"
                            "Loss {loss.val:.5f} ({loss.avg:.5f})\t"
                                .format(
                                epoch + 1,
                                self.trainer.max_epoch,
                                curr_step,
                                len(train_dataloader) * self.trainer.max_epoch,
                                loss=losses,
                        )
                    )

                if curr_step % self.trainer.eval_per_steps == 0 and val_dataloader is not None:
                    best_metrics = self.val_step(val_dataloader, evaluators, checkpoint_path, best_metrics)
                    if self.early_stop_cur == 0:
                        return True

        return True if best_metrics!={} else False


    @torch.no_grad()
    def inference_step(self,
                   test_dataloader: DataLoader,
                   task_name: str,
                   )-> List[numpy.ndarray]:

        self.realnet.eval()
        self.backbone.eval()

        preds = []
        with torch.no_grad():
            for i, input in enumerate(test_dataloader):
                outputs = self.realnet_emb(input, train=False)
                outputs.update(input)
                outputs = self.realnet(outputs, train = False)
                preds.append(outputs["anomaly_score"])
        preds = torch.cat(preds, dim=0).cpu().numpy()
        assert preds.shape[0] == len(test_dataloader.dataset)
        preds = numpy.squeeze(preds, axis=1)
        if task_name != 'val':
            preds = self.patch_post_processing(preds)
        preds = [pred for pred in preds]  # N x H x W
        return preds

    def save_checkpoint(self, checkpoint_path: str):
        torch.save({'state_dict': self.realnet.state_dict(),
                    'max_anomaly_score': self.max_anomaly_score,
                    'min_anomaly_score': self.min_anomaly_score,
                    'fusion_weights': self.fusion_weights,
        }, checkpoint_path)


    def load_checkpoint(self, checkpoint_path: str):
        state_dict = torch.load(checkpoint_path, map_location=self.device)
        self.realnet.load_state_dict(state_dict['state_dict'])
        self.min_anomaly_score = state_dict['min_anomaly_score']
        self.max_anomaly_score = state_dict['max_anomaly_score']
        self.fusion_weights = state_dict['fusion_weights'] if 'fusion_weights' in state_dict else None

